/*
 * A pipe paging coordinator, permitting memory sharing pipe endpoints.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <mem/memory_preallocated.h>

#include "page_queue_partitioned.h"
#include "pipe_paging.h"

#include "notifier_client.h"



PipePaging::PipePaging(Memory *memory, offset_t size)
: _memory(NULL), _size(size)
{
    /* Reserve space for two pipe regions. */

    _memory = new MemoryPreallocated(memory, round(size, memory->region_size()) * 2);
    _queue = new PageQueuePartitioned();
    _pages = new Pages(_memory, _queue);

    /* Reset the mappers associated with the regions. */

    for (unsigned int i = 0; i < 2; i++)
        _regions[i] = NULL;

    /* Initialise endpoints and flags for notifications. */

    for (unsigned int i = 0; i < 2; i++)
    {
        _flags[i] = 0;
        _deferred[i] = 0;
    }
}

/* Subscribe to an endpoint's notifications using a notification endpoint. */

void PipePaging::subscribe(bool writing, l4_cap_idx_t notifier, notify_flags_t flags)
{
    int i = writing ? 1 : 0;

    _notifiers[i].insert(notifier);
    _flags[i] = flags;

    /* Send deferred conditions on behalf of the other endpoint held before
       subscription occurred. */

    if (_deferred[i])
    {
        notify(!writing, _deferred[i]);
        _deferred[i] = 0;
    }
}

/* Unsubscribe from an endpoint's notifications. */

void PipePaging::unsubscribe(bool writing, l4_cap_idx_t notifier)
{
    int i = writing ? 1 : 0;
    std::set<l4_cap_idx_t>::iterator it = _notifiers[i].find(notifier);

    if (it != _notifiers[i].end())
    {
        _notifiers[i].erase(it);
        _flags[i] = 0;
        _deferred[i] = 0;
        ipc_cap_free_um(notifier);
    }
}

/* Notify the other endpoint. */

void PipePaging::notify(bool writing, notify_flags_t flags)
{
    /* Let the writer notify the reader, and the other way round. */

    int i = writing ? 0 : 1;

    /* Notify the other endpoint or hold any notification for potential future
       subscription. */

    if (!_notifiers[i].empty())
    {
        if (flags & _flags[i])
        {
            std::set<l4_cap_idx_t>::iterator it;

            for (it = _notifiers[i].begin(); it != _notifiers[i].end(); it++)
            {
                client_Notifier notifier(*it);

                notifier.notify(flags & _flags[i]);
            }
        }
    }
    else
        _deferred[i] = flags;
}

/* Return whether one or more endpoints have detached. */

int PipePaging::closed()
{
    return _endpoints < 2;
}

void PipePaging::discard_region(unsigned int i)
{
    PageMapper *mapper = _regions[i];

    if (mapper != NULL)
    {
        mapper->detach();
        _regions[i] = NULL;
        delete mapper;
    }
}

/* Detach one endpoint. */

void PipePaging::detach()
{
    if (!_endpoints)
        return;
    else
        _endpoints--;

    /* Return if the other endpoint is attached. */

    if (_endpoints)
        return;

    /* Discard all regions from the pipe. */

    for (unsigned int i = 0; i < 2; i++)
        discard_region(i);

    /* Release notifiers. */

    for (unsigned int i = 0; i < 2; i++)
    {
        std::set<l4_cap_idx_t>::iterator it;

        for (it = _notifiers[i].begin(); it != _notifiers[i].end(); it++)
            ipc_cap_free_um(*it);

        _notifiers[i].clear();
    }

    /* Delete the page collection and related objects. */

    delete _pages;
    delete _queue;
    delete _memory;
}

/* Add a region to the sequence. */

PageMapper *PipePaging::add_region()
{
    /* If the writer already accesses a different region to the reader, no new
       region is added. */

    if (_writing != _reading)
        return NULL;

    /* Select the other region of the pair being maintained. */

    _writing = 1 - _writing;

    /* Make a new mapper for the region. */

    PageMapper *mapper = new PageMapper(&_accessors[_writing], _pages);

    /* Initialise and record the mapper. */

    mapper->attach();
    mapper->set_data_size(0);

    _regions[_writing] = mapper;

    /* Let the writer notify the reader. */

    notify(true, NOTIFY_CONTENT_AVAILABLE);

    /* Return the next region's mapper. */

    return mapper;
}

/* Return the current region for reading. */

PageMapper *PipePaging::current_region()
{
    return _regions[_reading];
}

/* Return the next region for the reader if the writer is using a different one.
   Otherwise, return NULL. */

PageMapper *PipePaging::next_region()
{
    /* If the reader already accesses the same region to the writer, no next
       region can be obtained. */

    if (_reading == _writing)
        return NULL;

    /* Detach and discard the current page mapper. */

    discard_region(_reading);

    /* Select the next region. */

    _reading = 1 - _reading;

    /* Let the reader notify the writer. */

    notify(false, NOTIFY_SPACE_AVAILABLE);

    /* Return the next region's mapper. */

    return _regions[_reading];
}

// vim: tabstop=4 expandtab shiftwidth=4
